#!/usr/bin/perl
#
# This PERL script computes the probability of scarring for
#   all trees in the FHX2 fire history file.
#
# This file MUST have the following special fhx2 format:
#	The fire history records are stored in columns, one col. per tree.
#	Multiple trees are collected in groups --  a single plot.
#	The header portion of the file gives the site, plot, and tree ID's as:
#		First [P] rows give plot ID
#		Next  [T] rows give tree ID within plot
#       The numbers in [brackets] can be set with the constants below:
$pIDsize = 6;
$tIDsize = 1;
#	NOTE: pIDsize + tIDsize MUST EQUAL # hdr rows!!!
# e.g.
#	FHX2 FORMAT	- first record identifies FHX2 file format.
# 	1527 5 7	- first year in record, # trees (cols), # hdr rows
# 	DDDDD		|
#	CCCCC		 \
#	RRRRR		  \_ [six] character plot ID
#	00000		  /
#	11111		 /
#	55566		| 
#	12312		- [one] character tree ID within plot
#			- blank row 
#	.....		\
#	.....		- data rows in FHX2 format...
#	.....		/
#
# The list of symbols recognized as indicating a fire scar are:
$ScarSymb = 'ADEMLUd';
# The list of symbols that indicates a tree is a recording tree are:
$StartSymb = $ScarSymb.'|';
# The list of symbols that indicates a tree is finished recording are:
$EndSymb = '}]';
#
# 
# It will produce a data table in the following format:
#
# Plot#  treeList sYear eYear #Years #Trees #RecFire CtList  N^   p^   Gt
# (ID)   (ID list)  (period)  (e-s+1)  (t)    (Mt)  (C1,..Ct) (see below)
#-------|--------|-----|-----|------|------|-------|-------|---|-----|----
# DCR015  1,2     1680  1850    171    2      10     8, 6    12        .71
#                               171    2      10     8       12  .66
#                               171    2      10     6       12  .5  
# DCR015  1,3     1650  1815    165    2       8     8, 4    10        .83
#                               165    2       8     8       10  .8
#                               165    2       8     4       10  .4
# ...      ...
#
# Each primary record gives all details for the calculation (e.g. row 1).
# Each secondary record gives the p^ value each tree in the set.
#
# N^ = S(Ct * Mt-1)/S(Rt)   -- Mark Recapture estimate of N = # fires in #Years.
# p^ = Ct/N^                -- Mark Recapture estimate of p = prob. of scarring.
# Gt = Mt/S(Ct)             -- Emprical estimate of G(t) for curve fitting.
#
#-----------------------------------------------------------------------------
$ARGV[0] || die " usage: scarprob.pl fhx2File1 [fhx2File2] [...]\n";

#-----------------------------------------------------------------------------
# STEP 1: READ THE FHX2 FIRE HISTORY DATA INTO MEMORY ARRAYS
# For each file on the command line, read data:
#    Loop through all records.
#     - Read the first row to determine the header size, etc.
#     - Read the header rows for each tree to get site, plot & tree ID's.
#     - Read the fire history data for each tree.
#-----------------------------------------------------------------------------

foreach $file (@ARGV) {
  open(FHX, "<$file") || die "cannot open FHX2 file $file!\n";
  $row = -2;
  while (<FHX>) {
   $row++;
   chop;
#-----------------------------------------------------------------------------
#  - Read the first row to check the file format.
#    Bomb if the file is not formatted as expected.
#-----------------------------------------------------------------------------
   if ($row == -1) {
      @fields = split;
      die "File format is not FHX2!!\nFirst token = $fields[0]\n"
         if ($fields[0] != 'FHX2');
   }
#-----------------------------------------------------------------------------
#  - Read the second row to determine the header size, etc.
#    Bomb if the file is not formatted as expected.
#-----------------------------------------------------------------------------
   elsif ($row == 0) {
      @fields = split;
      $Year0 = $fields[0];   $year = $Year0;
      $Trees = $fields[1];
      $HdrRows = $fields[2];
      $tmp = $pIDsize + $tIDsize;
      die "Header rows, $HdrRows, does not equal internal header size, $tmp\n"
         if ($HdrRows != $tmp);
      for ($t=0; $t<$Trees; $t++) {$treeID[$t] = "";}
   }
#-----------------------------------------------------------------------------
#  - Read the header rows for each tree to get plot & tree ID's.
#  - There is a blank row after the header.
#-----------------------------------------------------------------------------
   elsif ($row <= $pIDsize) {
#  Plot ID's
      for ($t=0; $t<$Trees; $t++) {
          if (defined($pID[$t]))  {$pID[$t] .= substr($_,$t,1);}
          else {$pID[$t] = substr($_,$t,1);}
          }
   }
   elsif ($row <= $pIDsize+$tIDsize) {
#  Tree ID's
      for ($t=0; $t<$Trees; $t++) {
          if (defined($treeID[$t]))  {$treeID[$t] .= substr($_,$t,1);}
          else {$treeID[$t] = substr($_,$t,1);}
          }
   }
   elsif ($row == $HdrRows+1) {
#  Blank row (done reading header, so init data arrays).
      for ($t=0; $t<$Trees; $t++) {
         $TreeInd[$t] = "$pID[$t].$treeID[$t]";
#        Check that we haven't already seen this tree!!
         die "Tree $TreeInd[$t] is defined more than once!!!\n"
            if (defined($StartRec{$TreeInd[$t]}));
#        Save all the plot ID's in a seperate array.
         $PlotID{$pID[$t]} = 1;
         $StartRec{$TreeInd[$t]} = -1;
         $EndRec{$TreeInd[$t]} = -1;
         $FireYears{$TreeInd[$t]} = "";
      }
      undef @pID; undef @treeID;
   }

#-----------------------------------------------------------------------------
#  - Read the fire history data for each year.
#     - store the start & end record, and all fire years for each tree.
#     - keep track of all marker types found, for report.
#-----------------------------------------------------------------------------
   else {
      for ($t=0; $t<$Trees; $t++) {
         $marker = substr($_,$t,1);
         if ($StartRec{$TreeInd[$t]} == -1) {
            $StartRec{$TreeInd[$t]} = $year 
               if (index($StartSymb,$marker) >= 0);
         } else {
            $EndRec{$TreeInd[$t]} = $year 
               if (index($EndSymb,$marker) >= 0);
         }
         $FireYears{$TreeInd[$t]} .= "$year "
               if (index($ScarSymb,$marker) >= 0);

         $mrk{$marker} = 1;
      }
      $year++;
   }
  } # Next annual record...
#----------------------------


#-----------------------------------------------------------------------------
# Print a little report about the interpretations made above.
# YOU SHOULD CHECK THIS REPORT CAREFULLY when you run this script!!
#-----------------------------------------------------------------------------
   print(STDERR "\n\nReport for file $file ($Year0, $Trees, $HdrRows):\n");
   print(STDERR "Typical ID: Site.Plot.Tree = $TreeInd[0]\n");
   print(STDERR "Markers used to indicate fire scars: $ScarSymb\n");
   print(STDERR "Markers found in file: ");
   foreach $marker (keys(%mrk))	
	{print(STDERR "$marker ");}
   print(STDERR "\n\n");
   undef %mrk;
# Make sure each record at least had a start and end year.
  for ($t=0; $t<$Trees; $t++) {
     if ($StartRec{$TreeInd[$t]} == -1) {
        print (STDERR "Error: No start of record for tree $TreeInd[$t]\n");
        $StartRec{$TreeInd[$t]} = $year;
     }
     if ($EndRec{$TreeInd[$t]} == -1) {
        print (STDERR "Error: No end of record for tree $TreeInd[$t]\n");
        $EndRec{$TreeInd[$t]} = $Year0;
     }
     if (length($FireYears{$TreeInd[$t]}) == 0) 
        {print (STDERR "Warning: No fire years found for tree $TreeInd[$t]\n");}
   }

   close(FHX);
   undef @TreeInd;
} # Next file...
#----------------

#-----------------------------------------------------------------------------
# STEP 2: Compute Ct, Mt, and Rt for each pair, triple, quad, etc. of trees
#	  that recorded fires on the same site for the same time period.
# The algorithm works like this:
#    Gather up the indices of all trees in the same plot.
#    Determine all the possible tuples (pairs, triples, etc.)
#    Determine the period of overlap for each tuple.
#    Compute Ct, Mt, Rt, and Gt for each tuple as we go....
#-----------------------------------------------------------------------------
# MAIN LOOP over all PLOTS:
$First = 1;
plot: foreach $Plot (keys(%PlotID)) {
#print (STDERR "Plot=$Plot\n");
#  Get the indicies of all trees in this plot.
#@k=keys(%StartRec);
#print(STDERR "Keys=", join(" ", sort(@k)), "\n");
   @PlotTrees = grep(/$Plot\./,keys(%StartRec));
   $N = $#PlotTrees+1;
#print(STDERR "N=$N\n");
#  If there is only one tree in the plot, it is useless, make a note and skip.
   if ( $N <= 1) {
      print(STDERR "Plot $Plot has only $N trees & is not included in output!\n");
      next plot;
   }
#  Get a list of all possible tuples made up of the N trees in this plot.
   @tuples = &get_tuples($N);
#  LOOP over all tuples and estimate mark-recapture and G(t) values.
tuple: foreach $tuple (@tuples) {
#print STDERR "tuple = $tuple\n";
      @trees = split(' ',$tuple);
#     Determine the period of overlap for this tuple.
      $start = 0;  $end = 100000000000;
      foreach $tree (@trees) {
         $start = $StartRec{$PlotTrees[$tree]}
            if ($start < $StartRec{$PlotTrees[$tree]});
         $end = $EndRec{$PlotTrees[$tree]}
            if ($end > $EndRec{$PlotTrees[$tree]});
      }
#     Just skip this tuple if the period of overlap is less than one year.
      next tuple if ($end - $start < 1);

#     Compute Ct, Mt, and Rt for each tree in this tuple
      undef %UniqueFires;
      undef @C;
      $SumCt = 0;
      $SumCtMt = 0;
      $SumRt = 0;
      $Mt = 0;
tree: foreach $tree (@trees) {
         $C[$tree] = 0;
         @fires = split(' ',$FireYears{$PlotTrees[$tree]});
#        Check each fire year on tree is in period and cound it.
fire:    foreach $fire (@fires) {
            next fire if ($fire > $end || $fire < $start);
            $C[$tree]++;
            $UniqueFires{$fire} = 1;
         } # Next fire on this tree.

#        ********************************************
#        Compute mark-recapture values for this tree.
#        ********************************************
         $SumCt += $C[$tree];
         $SumCtMt += $C[$tree] * $Mt;
         @tmp = keys(%UniqueFires);
         $Mt = $#tmp + 1;
         $SumRt = $SumCt - $Mt;
      } # Next tree in this tuple

#     ********************************************
#     Compute estimates of #fires & G(t) (don't divide by 0!!)
#     ********************************************
      if ($SumRt > 0) {
         $Nfires = $SumCtMt/$SumRt;
#        Mark-Recapture can produce an N slightly smaller than Mt!!
         $Nfires = $Mt if ($Nfires<$Mt);
         $MR = 1;
      } else {
        print(STDERR "Tuple: ");
        foreach $tree (@trees) {print(STDERR "$PlotTrees[$tree],");}
        print(STDERR " had no duplicates -- not included in mark-recapture\n");
        $MR = 0;
      }
      $Gt = $Mt / $SumCt;

#-----------------------------------------------------------------------------
# STEP 3: Print data table record. 
#          - if this is the first record, print a little header...
#          - print a primary record for each tuple.
#          - print a secondary record for extra trees in each tuple.
#-----------------------------------------------------------------------------
     if ($First>0) {
      $First = 0;
      print("Plot#\ttreeList\tsYear\teYear\t#Years\t#Trees\t#RecFire\tCtList  ");
      print("\tN^\tp^\tGt\n");
      print("(ID)\t(ID list)\t(CommonPeriod)\t \t(e-s+1)\t(t)\t(Mt)\t(C1..Ct)");
      print("\tS(Ct*Mt)/S(Rt)\tCt/N^\tMt/S(Ct)\n");
     }
#    Primary record:
     print("$Plot\t");
     foreach $tree (@trees) { 
        $treeID = substr($PlotTrees[$tree],rindex($PlotTrees[$tree],'.')+1);
        print("$treeID ");
     }
     $Ntrees = $#trees+1;
     $Nyears = $end-$start+1;
     print("\t$start\t$end\t$Nyears\t$Ntrees\t$Mt\t");
     foreach $tree (@trees) {print("$C[$tree] ");}
     if ($MR > 0) {
        print("\t$Nfires\t ");
     } else {
        print("\t \t ");
     }
     print("\t$Gt\n");

#    Secondary records (only if mark-recapture was computed):
     if ($MR > 0) {
        foreach $tree (@trees) {
          $p = $C[$tree]/$Nfires;
          print(" \t \t \t \t$Nyears\t$Ntrees\t$Mt\t$C[$tree]\t$Nfires\t$p\t \n");
        }
     }

   } # Next tuple for this plot.
} # Next plot.


#-----------------------------------------------------------------------------
# SUBROUTINE get_tuples:
#     - compute all possible 2,3,..N-tuples of N trees and 
#       return then as array of strings.
#-----------------------------------------------------------------------------
sub get_tuples
{  local ($N)=@_;
#  For each possible tuple size (between a pair and an N-1 tuple):
#     enumerate all possible tuples of that size
   local(@tuples, $TupSize, @TupInd, $tup);
   for ($TupSize = 2; $TupSize < $N; $TupSize++) {
#print(STDERR "Tuple Size = $TupSize, N = $N\n");
#     Initialize tuple as (0,1,2,...,TupSize-1)
      for ($i=0; $i<$TupSize; $i++) {$TupInd[$i] = $i;}
#     Enumerate all possible tuples of this size by counting with TupInd
#     Stop when TupInd[0], which starts at 0, reaches N-1.
      while ($TupInd[0] < $N-1) {
#        If this TupInd forms a valid tuple, push it onto the return array.
         if (&valid_tuple(@TupInd)>0) {
            $tup = join(' ',@TupInd);
#print(STDERR "Valid tuple found: $tup\n");
            push(@tuples,$tup);
         }
#        Generate the next tuple (start by incrementing leftmost counter).
         $i = $TupSize-1;
         while (++$TupInd[$i] >= $N) {
            $TupInd[$i] = $TupInd[$i-1]+1;
            $i--;
            die ("BUG -- This should never happen!!\n") if ($i < 0);
         }
      } # Next tuple of this size.
   } # Next sized tuple.
#  Finally, append the single N-tuple.
   $tup = '0';
   for ($i=1; $i<$N; $i++) {$tup .= " $i";}
   push(@tuples, $tup);
#print(STDERR "Returning @tuples\n");
   return(@tuples);
}

#-----------------------------------------------------------------------------
# SUBROUTINE valid_tuple:
#     - determines if a tuple is valid (all indicies must be unique).
#-----------------------------------------------------------------------------
sub valid_tuple
{  local (@tuple)=@_;
   local(%unique);
   foreach $ind (@tuple) {
      return(0) if ($unique{$ind} > 0);
      $unique{$ind} = 1;
   }
   return(1);
}
            
